unit ZClasses;

{Alternative class library
 Part of the z4free library package.

 For use with the Free Pascal Compiler (http://www.freepascal.org/) or the
 Lazarus project (http://www.lazarus.freepascal.org/)

 Copyright (C) 2007 Robert H. Mangl

 ZClasses is free software: you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
 Foundation, either version 3 of the License, or (at your option) any later
 version.

 ZClasses is distributed in the hope that it will be useful, but WITHOUT ANY
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with ZClasses.  If not, see <http://www.gnu.org/licenses/>.}
 
 {$mode objfpc}{$H+}

 interface

  uses Classes, Math, SysUtils,
       Zett;
  
  const

   // Type Status Flags
   tfNamed             = 0; // data instance has a name
   tfHasChanged        = 1; // data has changed somehow
   tfOwnsData          = 2; // tells if vector should treat data as own
   tfReversed          = 3; // reversed order of sequence
   tfDuplicates        = 4; // allows duplicate keys
   tfFound             = 5; // item found
   tfNext              = 6; // uses this identifier as start for your flags
   
   // Find Status Flags
   fsNoFind = -1;
   
   // Name Status Flags
   nsInexistent = -1; // name not found at all
   nsAmbiguous  = -2; // name found at least twice

  type
   TData   = class;
   TVector = class;

   TType = class( TObject)
     private
      constructor Init;
     protected
      procedure ReadFrom( S: TStream); virtual; abstract;
      procedure WriteTo( S: TStream); virtual; abstract;
    end;
    
   TTypeClass = class of TType;
   
   { TBitAry }

   TBitAry = class( TType)
     private
      fBits: array of Cardinal;
      fCount: Cardinal;
      function GetAsString: string;
      function GetBit( Index: Cardinal): Boolean;
      function GetHigh: Cardinal;
      procedure SetAsString( Value: string);
      procedure Split( Index: Integer; out ArrayIndex: SizeInt;
                                       out BitValue: Cardinal);
      procedure SetBit( Index: Cardinal; Value: Boolean);
     protected
      procedure ReadFrom( S: TStream); override;
      procedure WriteTo( S: TStream); override;
     public
      constructor Create;
      destructor Destroy; override;
      property AsString: string read GetAsString write SetAsString;
      property Bit[ Index: Cardinal]: Boolean read GetBit write SetBit; default;
      property Count: Cardinal read fCount;
      property HighestBit: Cardinal read GetHigh;
      procedure CopyFrom( Original: TBitAry);
      procedure Pack;
    end;

   { TData }

   TData = class( TType)
     private
      fFlags: TBitAry;
      fName: string;
      fParent: TVector;
      fOnChange: TNotifyEvent;
      fOnNameChange: TNotifyEvent;
      function GetFlag( Index: Cardinal): Boolean;
     protected
      procedure ReadFrom( S: TStream); override;
      procedure WriteTo( S: TStream); override;
      function CompareWith( aKey: Pointer): Integer; virtual;
      function GetHasChanged: Boolean;
      function GetKey: Pointer; virtual;
      function GetName: string; virtual;
      procedure NotifyChange;
      procedure NotifyNameChange;
      procedure SetFlag( Index: Cardinal; Value: Boolean); virtual;
      procedure SetHasChanged( Value: Boolean);
      procedure SetName( Value: string); virtual;
     public
      constructor Create;
      constructor Create( aName: string);
      destructor Destroy; override;
      property Flag[ Index: Cardinal]: Boolean read GetFlag write SetFlag;
      property HasChanged: Boolean read GetHasChanged write SetHasChanged;
      property Key: Pointer read GetKey;
      property Name: string read GetName write SetName;
      function NextSibling: TData;
      property OnChange: TNotifyEvent read fOnChange write fOnChange;
      property OnNameChange: TNotifyEvent read fOnNameChange
                                          write fOnNameChange;
      property Parent: TVector read fParent;
      function PrevSibling: TData;
      function OwnIndex: SizeInt;
    end;
    
   TFloat = class( TData)
     private
      fValue: Float;
     public
      constructor Create( aValue: Float);
      property Value: Float read fValue write fValue;
    end;
    
   TItem = class( TData)
     private
      fHandle: Cardinal;
      fRefCount: Integer;
     protected
      procedure ReadFrom( S: TStream); override;
      procedure WriteTo( S: TStream); override;
      function CompareWith( aKey: Pointer): Integer; override;
      function GetKey: Pointer; override;
     public
      constructor Create;
      constructor Create( aName: string);
      property Handle: Cardinal read fHandle;
      function IsLocked: Boolean;
      procedure Lock;
      procedure Permalock;
      procedure Unlock;
    end;
    
   { TVector }

   TVector = class( TData)
     private
      fData: array of TData;
      fCount: Cardinal;
      fDelta, fLimit: SizeInt;
      procedure Confirm( Index: SizeInt; WantInsert: Boolean);
      procedure ExpandTo( Index: SizeInt);
      function GetData( Index: SizeInt): TData;
      function GetMaxIndex: SizeInt;
      procedure SetDelta( NewDelta: SizeInt);
      procedure SetLimit( NewLimit: SizeInt);
     protected
      procedure ReadFrom( S: TStream); override;
      procedure WriteTo( S: TStream); override;
      procedure HandleDataChange( Sender: TObject); virtual;
      procedure BeforeInsertion( aData: TData; Index: SizeInt); virtual;
      procedure AfterInsertion( aData: TData; Index: SizeInt); virtual;
      procedure BeforeRemoval( aData: TData; Index: SizeInt); virtual;
      procedure AfterRemoval( aData: TData; Index: SizeInt); virtual;
      function CanInsert( aData: TData; Index: SizeInt): Boolean; virtual;
      function CanRemove( aData: TData; Index: SizeInt): Boolean; virtual;
      procedure Error( ErrMsg: string); virtual;
      function PlaceFor( aData: TData): SizeInt; virtual;
     public
      constructor Create;
      constructor Create( aName: string);
      destructor Destroy; override;
      property Count: Cardinal read fCount;
      property Data[ Index: SizeInt]: TData read GetData;
      function DataByName( aName: string): TData;
      property Delta: SizeInt read fDelta write SetDelta;
      procedure FreeData( aData: TData);
      procedure FreeAll;
      procedure FreeAt( Index: SizeInt);
      function IndexByName( aName: string): SizeInt;
      function IndexOf( aData: TData): SizeInt; virtual;
      function InRange( Index: SizeInt): Boolean;
      function Insert( aData: TData): SizeInt;
      procedure InsertAt( Index: SizeInt; aData: TData);
      property Limit: SizeInt read fLimit write SetLimit;
      property MaxIndex: SizeInt read GetMaxIndex;
      procedure MoveData( Source, Dest: SizeInt);
      function Remove( aData: TData): TData;
      function RemoveAt( Index: SizeInt): TData;
      function Replace( OldData, NewData: TData): SizeInt;
      function ReplaceAt( Index: SizeInt; NewData: TData): TData;
      procedure Swap( Index1, Index2: SizeInt);
    end;
    
   TStack = class( TVector) // stack operations according to forth tradition
     protected
      procedure Error( ErrMsg: string); override;
     public
      constructor Create;
      function IsEmpty: Boolean;
      procedure Pop; // remove ToS and destroy it ( n - )
      function Pull: TData; // remove ToS and give it back: ( n - )
      procedure Push( aData: TData); // put element on the stack: ( - n)
      procedure Rotate; // rotate the last three elements: (n1 n2 n3 - n2 n3 n1)
      procedure Swap; // swap last two elements: (n1 n2 - n2 n1)
      function ToS: TData; // show ToS, don't remove it ( - )
    end;

   TQueue = class( TVector)
     public
      constructor Create;
      constructor Create( aName: string);
      procedure Enqueue( aData: TData);
      function Dequeue: TData;
      function FirstData: TData;
      function LastData: TData;
      function IsEmpty: Boolean;
    end;
    
   TFloatStack = class( TStack)
     public
      function PullVal: Float;
      procedure PushVal( Value: Float);
      function TopValue: Float;
    end;

   TSequence = class( TVector)
     private
      function HasDuplicates: Boolean;
     protected
      procedure ReadFrom( S: TStream); override;
      function PlaceFor( aData: TData): SizeInt; override;
      function SearchSub( Lo, Hi: SizeInt; aKey: Pointer): Integer;
      procedure SetFlag( Index: Cardinal; Value: Boolean); override;
     public
      function Compare( Index: SizeInt; aKey: Pointer): Integer;
      function IndexOf( aData: TData): SizeInt; override;
      function Search( aKey: Pointer): SizeInt;
      procedure Reverse;
      procedure Sort; virtual;
    end;
    
   { TPool }

   TPool = class( TSequence)
     private
      IsBeingDestroyed: Boolean;
      function GetItem( aHandle: Cardinal): TItem;
     protected
      procedure AfterRemoval( aData: TData; Index: SizeInt); override;
      procedure BeforeInsertion( aData: TData; Index: SizeInt); override;
      function CanInsert( aData: TData; Index: SizeInt): Boolean; override;
      function CanRemove( aData: TData; Index: SizeInt): Boolean; override;
      function PlaceFor( aData: TData): SizeInt; override;
     public
      constructor Create;
      constructor Create( aName: string);
      destructor Destroy; override;
      property Item[ aHandle: Cardinal]: TItem read GetItem;
      function ItemByName( aName: string): TItem;
      procedure Realloc( var Container: Cardinal; NewHandle: Cardinal);
    end;

   EBitAry = class( EZett);
   EVector = class( EZett);
   EZtream = class( EZett);
   EStack  = class( EZett);

  procedure RegisterType( aTypeClass: TTypeClass);
  procedure RegisterZClassesClasses;
  
  function ReadCardinal( S: TStream): Cardinal;
  function ReadFloat( S: TStream): Float;
  function ReadInteger( S: TStream): Integer;
  function ReadSizeInt( S: TStream): SizeInt;
  function ReadType( aTypeClass: TTypeClass; S: TStream;
                     DataParent: TVector = nil): TType;

  procedure WriteCardinal( aCardinal: Cardinal; S: TStream);
  procedure WriteFloat( aFloat: Float; S: TStream);
  procedure WriteInteger( anInteger: Integer; S: TStream);
  procedure WriteSizeInt( aSizeInt: SizeInt; S: TStream);
  procedure WriteType( aType: TType; S: TStream);

 implementation
 
  const
   CardinalWidth = 8 * SizeOf( Cardinal);

  type
   TStreamEntry = class( TData)
     fTypeClass: TTypeClass;
     fClassName: string;
     constructor Create( aTypeClass: TTypeClass);
     function CompareWith( aKey: Pointer): Integer; override;
     function CreateType: TType;
     function GetKey: Pointer; override;
    end;

   { TStrItem }

   TStrItem = class( TData)
     private
      fContent: string;
     public
      constructor Create( aName, aContent: string);
      property Content: string read fContent write fContent;
    end;

  var
   Registry      : TSequence  = nil;
   LastTypeClass : TTypeClass = nil;

  // TBitAry methods

   constructor TBitAry.Create;
    begin
     inherited Create;
     SetLength( fBits, 0);
     fCount := 0
    end;
    
   destructor TBitAry.Destroy;
    begin
     SetLength( fBits, 0);
     inherited Destroy
    end;
    
   procedure TBitAry.ReadFrom( S: TStream);
    var
     L: SizeInt;
    begin
     L := ReadSizeInt( S);
     SetLength( fBits, L);
     if L <> 0
      then S.ReadBuffer( fBits[ 0], L * SizeOf( Cardinal))
    end;

   procedure TBitAry.WriteTo( S: TStream);
    var
     L: SizeInt;
    begin
     Pack;
     L := High( fBits) + 1;
     WriteSizeInt( L, S);
     if L <> 0
      then S.WriteBuffer( fBits[ 0], L * SizeOf( Cardinal))
    end;

   procedure TBitAry.CopyFrom( Original: TBitAry);
    var
     OriginalSize: Integer;
    begin
     OriginalSize := High( Original.fBits) + 1;
     if OriginalSize > 0
      then fBits := Copy( Original.fBits, 0, OriginalSize)
      else SetLength( Original.fBits, 0)
    end;

   function TBitAry.GetAsString: string;
    var
     S: string;
     L: SizeInt;
    begin
      Pack;
      L := Length( fBits) * 4;
      if L = 0
       then S := ''
       else begin
         SetLength( S, L);
         Move( fBits, S, L)
        end;
      GetAsString := S
     end;

   function TBitAry.GetBit( Index: Cardinal): Boolean;
    var
     ArrayIndex: SizeInt;
     BitValue: Cardinal;
    begin
     Split( Index, ArrayIndex, BitValue);
     if ArrayIndex > High( fBits)
      then GetBit := False
      else GetBit := (fBits[ ArrayIndex] and BitValue) <> 0
    end;

   function TBitAry.GetHigh: Cardinal;
    var
     L, I: SizeInt;
     B, H: Cardinal;
    begin
      if Count = 0
       then Exit( 0);
      Pack;
      L := Length( fBits);
      H := fBits[ L];
      I := MSBCardIndex;
      B := MSBCardinal;
      while (H and B) = 0 do
       begin
         B := B shr 1;
         Dec( I)
        end;
      GetHigh := SizeOf( Cardinal) * 8 * (L - 1) + I
     end;

   procedure TBitAry.SetAsString( Value: string);
    var
     L: SizeInt;
    begin
      if Value = ''
       then SetLength( fBits, 0)
       else begin
         L := (Length( Value) - 1) div SizeOf( Cardinal) + 1;
         SetLength( fBits, L);
         Move( Value, fBits, L)
        end;
     end;

   procedure TBitAry.Pack;
    var
     OldHigh, NewHigh: Integer;
    begin
     OldHigh := High( fBits);
     NewHigh := OldHigh;
     while (NewHigh >= 0) and (fBits[ NewHigh] = 0) do
      Dec( NewHigh);
     if NewHigh <> OldHigh
      then SetLength( fBits, NewHigh + 1)
    end;

   procedure TBitAry.Split( Index: Integer; out ArrayIndex: SizeInt; out
    BitValue: Cardinal);
    var
     BitIndex: Cardinal;
    begin
     if Index < 0
      then raise EBitAry.Create( 'bit index less than 0');
     ArrayIndex := Index div CardinalWidth;
     BitIndex := Index mod CardinalWidth;
     BitValue := 1 shl BitIndex
    end;

   procedure TBitAry.SetBit( Index: Cardinal; Value: Boolean);
    var
     ArrayIndex: SizeInt;
     BitValue: Cardinal;
     OldValue: Boolean;
    begin
     OldValue := GetBit( Index);
     if OldValue = Value
      then Exit;
     Inc( fCount, 2 * Ord( Value) - 1);
     Split( Index, ArrayIndex, BitValue);
     if Value and (ArrayIndex > High( fBits))
      then SetLength( fBits, ArrayIndex + 1);
     if Value
      then fBits[ ArrayIndex] := fBits[ ArrayIndex] or BitValue
      else if ArrayIndex <= High( fBits)
            then fBits[ ArrayIndex] := fBits[ ArrayIndex] and not BitValue;
    end;

  // TData methods

   constructor TData.Create;
    begin
     inherited Create;
     fName := '';
     fFlags := TBitAry.Create;
     fParent := nil;
     fOnChange := nil
    end;

   constructor TData.Create( aName: string);
    begin
     Create;
     fName := aName;
     Flag[ tfNamed] := fName <> ''
    end;
    
   destructor TData.Destroy;
    begin
     fName := '';
     fFlags.Free;
     inherited Destroy
    end;

   procedure TData.ReadFrom( S: TStream);
    begin
     fFlags := TBitAry( ReadType( TBitAry, S));
     Flag[ tfHasChanged] := False;
     if Flag[ tfNamed]
      then fName := S.ReadANSIString
      else fName := ''
    end;

   procedure TData.WriteTo( S: TStream);
    begin
     HasChanged := False;
     WriteType( fFlags, S);
     if Flag[ tfNamed]
      then S.WriteANSIString( fName);
    end;

   function TData.CompareWith( aKey: Pointer): Integer;
    begin
     CompareWith := ANSICompareText( Name, PString( aKey)^)
    end;

   function TData.GetFlag( Index: Cardinal): Boolean;
    begin
     GetFlag := fFlags[ Index]
    end;

   function TData.GetHasChanged: Boolean;
    begin
     GetHasChanged := Flag[ tfHasChanged]
    end;

   function TData.GetKey: Pointer;
    begin
     GetKey := @fName
    end;

   function TData.GetName: string;
    begin
      GetName := fName
     end;

   function TData.NextSibling: TData;
    var
     I: Integer;
    begin
      if not Assigned( Parent)
       then Exit( nil);
      I := Parent.IndexOf( Self);
      if I >= Parent.MaxIndex
       then Exit( nil);
      NextSibling := Parent.Data[ I + 1]
     end;
    
   procedure TData.NotifyChange;
    begin
     if Assigned( fOnChange)
      then OnChange( Self)
    end;

   procedure TData.NotifyNameChange;
    begin
      HasChanged := True;
      if Assigned( fOnNameChange)
       then OnNameChange( Self)
     end;

   function TData.OwnIndex: SizeInt;
    begin
      if Assigned( Parent)
       then OwnIndex := Parent.IndexOf( Self)
       else OwnIndex := -1
     end;

   function TData.PrevSibling: TData;
    var
     I: Integer;
    begin
      if not Assigned( Parent)
       then Exit( nil);
      I := Parent.IndexOf( Self);
      if I <= 0
       then Exit( nil);
      PrevSibling := Parent.Data[ I - 1]
     end;

   procedure TData.SetFlag( Index: Cardinal; Value: Boolean);
    begin
     if fFlags[ Index] <> Value
      then fFlags[ Index] := Value
    end;
     
   procedure TData.SetHasChanged( Value: Boolean);
    begin
     Flag[ tfHasChanged] := Value;
     if Value
      then NotifyChange
    end;

   procedure TData.SetName( Value: string);
    begin
     if fName = Value
      then Exit;
     fFlags[ tfNamed] := Value <> '';
     fName := Value;
     NotifyNameChange
    end;

  // TFloat methods
       
   constructor TFloat.Create( aValue: Float);
    begin
     inherited Create;
     fValue := aValue
    end;
    
  // TFloatStack methods
      
   function TFloatStack.PullVal: Float;
    var
     F: TFloat;
    begin
     F := TFloat( Pull);
     PullVal := F.Value;
     F.Free
    end;

   procedure TFloatStack.PushVal( Value: Float);
    begin
     Push( TFloat.Create( Value))
    end;

   function TFloatStack.TopValue: Float;
    var
     aFloat: TFloat;
    begin
     aFloat := TFloat( ToS);
     if aFloat = nil
      then Error( 'stack empty')
      else TopValue := aFloat.Value
    end;
    
  // TItem methods
       
   constructor TItem.Create;
    begin
     inherited Create;
     fHandle := 0;
     fRefCount := 0
    end;

   constructor TItem.Create( aName: string);
    begin
     inherited Create( aName);
     fHandle := 0;
     fRefCount := 0
    end;

   procedure TItem.ReadFrom( S: TStream);
    begin
     inherited ReadFrom( S);
     fHandle := ReadCardinal( S);
     fRefCount := ReadSizeInt( S)
    end;
    
   procedure TItem.WriteTo( S: TStream);
    begin
     inherited WriteTo( S);
     WriteCardinal( fHandle, S);
     WriteSizeInt( fRefCount, S)
    end;

   function TItem.CompareWith( aKey: Pointer): Integer;
    begin
     CompareWith := fHandle - PCardinal( aKey)^
    end;
    
   function TItem.GetKey: Pointer;
    begin
     GetKey := @fHandle
    end;
    
   function TItem.IsLocked: Boolean;
    begin
     IsLocked := fRefCount <> 0
    end;

   procedure TItem.Lock;
    begin
     if fRefCount <> -1
      then Inc( fRefCount)
    end;

   procedure TItem.Permalock;
    begin
     fRefCount := -1
    end;

   procedure TItem.Unlock;
    begin
     if fRefCount <> -1
      then Dec( fRefCount)
    end;

  // TPool methods
  
   constructor TPool.Create;
    begin
     inherited Create;
     IsBeingDestroyed := False;
     Flag[ tfOwnsData] := True
    end;
    
   constructor TPool.Create( aName: string);
    begin
     inherited Create;
     Name := aName
    end;

   destructor TPool.Destroy;
    begin
     IsBeingDestroyed := True;
     inherited Destroy
    end;
    
   procedure TPool.BeforeInsertion( aData: TData; Index: SizeInt);
    var
     NewItem: TItem;
    begin
     inherited BeforeInsertion( aData, Index);
     NewItem := TItem( aData);
     NewItem.fHandle := Index + 1
    end;
    
   procedure TPool.AfterRemoval( aData: TData; Index: SizeInt);
    begin
    TItem( aData).fHandle := 0;
     inherited AfterRemoval( aData, Index)
    end;

   function TPool.CanInsert( aData: TData; Index: SizeInt): Boolean;
    begin
     with TItem( aData) do
      CanInsert := not IsLocked and (Handle = 0)
    end;
    
   function TPool.CanRemove( aData: TData; Index: SizeInt): Boolean;
    begin
     with TItem( aData) do
      CanRemove := IsBeingDestroyed or not IsLocked
    end;

   function TPool.GetItem( aHandle: Cardinal): TItem;
    var
     Index: SizeInt;
    begin
     Index := Search( @aHandle);
     if Flag[ tfFound]
      then GetItem := TItem( Data[ Index])
      else GetItem := nil
    end;

   function TPool.ItemByName( aName: string): TItem;
    var
     Index: SizeInt;
    begin
     Index := IndexByName( aName);
     if Index = -1
      then ItemByName := nil
      else ItemByName := TItem( Data[ Index]);
    end;

   function TPool.PlaceFor( aData: TData): SizeInt;
    var
     Lo, Hi, Index: SizeInt;
    begin
     Lo := 0;
     Hi := MaxIndex;
     while Lo <= Hi do
      begin
       Index := (Lo + Hi) div 2;
       if TItem( fData[ Index]).Handle = Index + 1
        then Lo := Index + 1
        else Hi := Index - 1
      end;
     PlaceFor := Lo
    end;
    
   procedure TPool.Realloc( var Container: Cardinal; NewHandle: Cardinal);
    var
     theItem: TItem;
    begin
     theItem := Item[ Container];
     if theItem <> nil
      then theItem.Unlock;
     Container := 0;
     theItem := Item[ Container];
     if theItem <> nil
      then theItem.Lock;
     Container := NewHandle
    end;
    
  // TQueue

   constructor TQueue.Create;
    begin
     inherited Create
    end;

   constructor TQueue.Create( aName: string);
    begin
     inherited Create( aName)
    end;

   function TQueue.Dequeue: TData;
    begin
     if not IsEmpty
      then Dequeue := RemoveAt( 0)
      else Dequeue := nil
    end;

   procedure TQueue.Enqueue( aData: TData);
    begin
     Insert( aData)
    end;

   function TQueue.FirstData: TData;
    begin
     if not IsEmpty
      then FirstData := Data[ 0]
      else FirstData := nil
    end;

   function TQueue.IsEmpty: Boolean;
    begin
     IsEmpty := Count = 0
    end;

   function TQueue.LastData: TData;
    begin
     if not IsEmpty
      then LastData := Data[ Count - 1]
      else LastData := nil
    end;

  // TSequence methods
       
   procedure TSequence.ReadFrom( S: TStream);
    begin
     inherited ReadFrom( S);
     Sort
    end;
    
   function TSequence.Compare( Index: SizeInt; aKey: Pointer): Integer;
    var
     CompResult, Direction: Integer;
    begin
     CompResult := fData[ Index].CompareWith( aKey);
     Direction := 1 - 2 * Ord( fFlags[ tfReversed]);
     Compare := Sign( CompResult) * Direction
    end;

   function TSequence.HasDuplicates: Boolean;
    var
     Index: Integer;
    begin
     if Count = 1
      then HasDuplicates := False
      else for Index := 0 to Count - 2 do
            if Compare( Index, Data[ Index + 1].Key) = 0
             then begin
               HasDuplicates := True;
               Break
              end
    end;
    
   function TSequence.IndexOf( aData: TData): SizeInt;
    var
     Index: SizeInt;
    begin
     Index := Search( aData.Key);
     if not fFlags[ tfFound]
      then Index := -1
      else if fFlags[ tfDuplicates]
            then repeat
                  Dec( Index);
                  if Index = -1
                   then begin
                    fFlags[ tfFound] := False;
                    Break
                   end
                 until fData[ Index] = aData;
     IndexOf := Index
    end;
    
   function TSequence.PlaceFor( aData: TData): SizeInt;
    var
     Place: SizeInt;
    begin
     Place := Search( aData.Key);
     if fFlags[ tfFound]
      then if fFlags[ tfDuplicates]
            then Inc( Place)
            else Error( 'no duplicates allowed');
     PlaceFor := Place
    end;
    
   function TSequence.Search( aKey: Pointer): SizeInt;
    begin
     Search := SearchSub( 0, MaxIndex, aKey)
    end;
    
   function TSequence.SearchSub( Lo, Hi: SizeInt; aKey: Pointer): SizeInt;
    var
     CmpRes: Integer;
     Index: SizeInt;
    begin
     fFlags[ tfFound] := False;
     while Lo <= Hi do
      begin
       Index := (Lo + Hi) div 2;
       CmpRes := Compare( Index, aKey);
       case CmpRes of
         -1: Lo := Index + 1;
         +1: Hi := Index - 1;
          0: begin
              fFlags[ tfFound] := True;
              if fFlags[ tfDuplicates]
               then Lo := Index + 1
               else begin
                     Hi := Index - 1;
                     Lo := Index
                    end
             end
        end
      end;
     if Flag[ tfFound] and Flag[ tfDuplicates]
      then SearchSub := Lo - 1
      else SearchSub := Lo
    end;

   procedure TSequence.SetFlag( Index: Cardinal; Value: Boolean);
    begin
     case Index of
       tfReversed  : if fFlags[ tfReversed] <> Value
                      then begin
                        Reverse;
                        Exit
                       end;
       tfDuplicates: if Value and HasDuplicates
                      then Error( 'duplicates found');
      end;
     inherited SetFlag( Index, Value)
    end;

   procedure TSequence.Sort;
    var
     Source, Target: SizeInt;
    begin
     if Count < 2
      then Exit;
     for Source := 1 to MaxIndex do
      begin
       if Compare( Source, Data[ Source - 1].Key) <> -1
        then begin
              Target := SearchSub( 0, Source - 1, fData[ Source].Key);
              MoveData( Source, Target)
             end
      end
    end;

   procedure TSequence.Reverse;
    var
     Index1, Index2, Last: SizeInt;
    begin
     if Count < 2
      then Exit;
     Last := Count div 2;
     if Count mod 2 = 1
      then Dec( Last);
     Index2 := Count - 1;
     for Index1 := 0 to Last do
      begin
       Swap( Index1, Index2);
       Dec( Index2)
      end;
     fFlags[ tfReversed] := not fFlags[ tfReversed]
    end;

  // TStack methods
  
   constructor TStack.Create;
    begin
     inherited Create;
     Flag[ tfOwnsData] := False
    end;
    
   procedure TStack.Error( ErrMsg: string);
   
    begin
     raise EStack.Create( ErrMsg)
    end;
    
   function TStack.IsEmpty: Boolean;
    begin
     IsEmpty := Count = 0
    end;
    
   procedure TStack.Pop;
    begin
     Pull.Free
    end;
    
   function TStack.Pull: TData;
    begin
     if Count = 0
      then Error( 'stack empty');
     Pull := ToS;
     RemoveAt( Count - 1)
    end;
    
   procedure TStack.Push( aData: TData);
    begin
     InsertAt( Count, aData)
    end;

   procedure TStack.Rotate;
    begin
     if Count < 3
      then Error( 'cannot rotate elements')
      else MoveData( Count - 3, Count - 1)
    end;
    
   procedure TStack.Swap;
    begin
     inherited Swap( Count - 1, Count - 2)
    end;
    
   function TStack.ToS: TData;
    begin
     if Count = 0
      then ToS := nil
      else ToS := Data[ Count - 1]
    end;

  // TStreamEntry methods

   constructor TStreamEntry.Create( aTypeClass: TTypeClass);
    begin
     inherited Create;
     fTypeClass := TTypeClass( aTypeClass.ClassType);
     fClassName := LowerCase( aTypeClass.ClassName)
    end;

   function TStreamEntry.CompareWith( aKey: Pointer): Integer;
    begin
     CompareWith :=
      CompareStr( UpperCase( fClassName), UpperCase( PString( aKey)^));
    end;

   function TStreamEntry.CreateType: TType;
    begin
     CreateType := fTypeClass.Init
    end;

   function TStreamEntry.GetKey: Pointer;
    begin
     GetKey := @fClassName
    end;
    
  // TStrItem

   constructor TStrItem.Create( aName, aContent: string);
    begin
     inherited Create( aName);
     fContent := aContent;
    end;

  // TType methods

   constructor TType.Init;
    begin
     inherited Create
    end;
    
  // TVector methods

   constructor TVector.Create;
    begin
     inherited Create;
     Flag[ tfOwnsData] := True;
     SetLength( fData, 0);
     fCount := 0;
     fDelta := 1;
     fLimit := MaxInt
    end;
    
   constructor TVector.Create( aName: string);
    begin
     Create;
     fName := aName;
     Flag[ tfNamed] := fName <> ''
    end;

   destructor TVector.Destroy;
    begin
     if Flag[ tfOwnsData]
      then FreeAll;
     inherited Destroy
    end;

   procedure TVector.ReadFrom( S: TStream);
    var
     Index: SizeInt;
     aData: TData;
    begin
     inherited ReadFrom( S);
     fCount := ReadCardinal( S);
     SetLength( fData, fCount);
     fDelta := ReadSizeInt( S);
     fLimit := ReadSizeInt( S);
     if fCount <> 0
      then for Index := 0 to MaxIndex do begin
        aData := TData( ReadType( nil, S, Self));
        if Flag[ tfOwnsData]
         then aData.OnChange := @HandleDataChange;
        fData[ Index] := aData
       end
    end;
    
   procedure TVector.WriteTo( S: TStream);
    var
     Index: SizeInt;
    begin
     inherited WriteTo( S);
     WriteCardinal( fCount, S);
     WriteSizeInt( fDelta, S);
     WriteSizeInt( fLimit, S);
     if fCount <> 0
      then for Index := 0 to MaxIndex do
            WriteType( fData[ Index], S)
    end;

   procedure TVector.AfterInsertion( aData: TData; Index: SizeInt);
    begin
    end;

   procedure TVector.AfterRemoval( aData: TData; Index: SizeInt);
    begin
     if Flag[ tfOwnsData]
      then begin
        aData.fParent := nil;
        aData.OnChange := nil
       end;
    end;

   procedure TVector.BeforeInsertion( aData: TData; Index: SizeInt);
    begin
     if Flag[ tfOwnsData]
      then begin
        aData.fParent := Self;
        aData.OnChange := @HandleDataChange
      end
    end;

   procedure TVector.BeforeRemoval(aData: TData; Index: SizeInt);
    begin
    end;

   function TVector.CanInsert( aData: TData; Index: SizeInt): Boolean;
    begin
     CanInsert := True
    end;

   function TVector.CanRemove( aData: TData; Index: SizeInt): Boolean;
    begin
     CanRemove := True
    end;
    
   procedure TVector.Confirm( Index: SizeInt; WantInsert: Boolean);
    var
     Failed: Boolean;
    begin
     if (Index < Count) or not WantInsert
      then Failed := not InRange( Index)
      else Failed := Index > Count;
     if Failed
      then Error( 'index ' + IntToStr( Index) + ' not in range [0 .. ' +
                  IntToStr( MaxIndex) + '] of ' + ClassName + ' ' + Name)
    end;

   function TVector.DataByName( aName: string): TData;
    var
     Index: SizeInt;
    begin
     Index := IndexByName( aName);
     if Index = -1
      then DataByName := nil
      else DataByName := Data[ Index]
    end;

   procedure TVector.Error( ErrMsg: string);
    begin
     raise EVector.Create( ErrMsg)
    end;
    
   procedure TVector.ExpandTo( Index: SizeInt);
    var
     CurHigh: SizeInt;
    begin
     if Index > Limit
      then Error( 'cannot expand over limit of ' + IntToStr( Limit));
     if Delta < 1
      then Error( 'expansion not allowed');
     CurHigh := High( fData);
     if Index > CurHigh
      then SetLength( fData, CurHigh + 1 + Delta);
     Inc( fCount)
    end;
    
   procedure TVector.FreeAll;
    var
     Index: Integer;
    begin
     while Count <> 0
      do FreeAt( Count - 1)
    end;

   procedure TVector.FreeAt( Index: SizeInt);
    var
     aData: TData;
    begin
     Confirm( Index, False);
     aData := fData[ Index];
     RemoveAt( Index);
     aData.Free
    end;
    
   procedure TVector.FreeData( aData: TData);
    begin
     FreeAt( IndexOf( aData))
    end;

   function TVector.GetMaxIndex: SizeInt;
    begin
     GetMaxIndex := Count - 1
    end;
    
   function TVector.GetData( Index: SizeInt): TData;
    begin
     Confirm( Index, False);
     GetData := fData[ Index]
    end;

   procedure TVector.HandleDataChange( Sender: TObject);
    begin
     HasChanged := True
    end;

   function TVector.IndexByName( aName: string): SizeInt;
    var
     I: SizeInt;
    begin
     IndexByName := -1;
     if Count = 0
      then Exit
      else for I := 0 to Count - 1 do
            if Data[ I].Name = aName
             then begin
                   IndexByName := I;
                   Exit
                  end;
                  
    end;

   function TVector.IndexOf( aData: TData): SizeInt;
    var
     I: SizeInt;
    begin
     IndexOf := -1;
     if Count > 0
      then for I := 0 to MaxIndex do
            if Data[ I] = aData
             then begin
                   IndexOf := I;
                   Break
                  end
    end;
    
   function TVector.InRange( Index: SizeInt): Boolean;
    begin
     InRange := (Index >= 0) and (Index < Count)
    end;
    
   function TVector.Insert( aData: TData): SizeInt;
    var
     Index: SizeInt;
    begin
     Index := PlaceFor( aData);
     InsertAt( Index, aData);
     Insert := Index
    end;
    
   procedure TVector.InsertAt( Index: SizeInt; aData: TData);
    var
     MoveSize: SizeInt;
    begin
     if IndexOf( aData) <> -1
      then Exit;
     Confirm( Index, True);
     if not CanInsert( aData, Index)
      then Error( 'cannot insert at ' + IntToStr( Index));
     BeforeInsertion( aData, Index);
     ExpandTo( Count);
     if Index < MaxIndex
      then begin
            MoveSize := (MaxIndex - Index) * SizeOf( TData);
            Move( fData[ Index], fData[ Index + 1], MoveSize)
           end;
     fData[ Index] := aData;
     HasChanged := True;
     AfterInsertion( aData, Index)
    end;
    
   procedure TVector.MoveData( Source, Dest: SizeInt);
    var
     MoveSize: SizeInt;
     Temp: TData;
    begin
     Confirm( Source, False);
     Confirm( Dest  , False);
     MoveSize := Abs( Source - Dest);
     if MoveSize = 1
      then begin
            Swap( Source, Dest);
            Exit
           end;
     if Source = Dest
      then Exit;
     MoveSize := MoveSize * SizeOf( TData);
     Temp := fData[ Source];
     if Dest > Source
      then Move( fData[ Source + 1], fData[ Source    ], MoveSize)
      else Move( fData[ Dest      ], fData[ Dest   + 1], MoveSize);
     fData[ Dest] := Temp
    end;
    
   function TVector.PlaceFor( aData: TData): SizeInt;
    begin
     PlaceFor := Count
    end;

   function TVector.Remove( aData: TData): TData;
    begin
     Remove := RemoveAt( IndexOf( aData))
    end;

   function TVector.RemoveAt( Index: SizeInt): TData;
    var
     aData: TData;
     MoveSize: Integer;
    begin
     Confirm( Index, False);
     aData := fData[ Index];
     if not CanRemove( aData, Index)
      then Error( 'custom removal error at ' + IntToStr( Index));
     BeforeRemoval( aData, Index);
     if Index < MaxIndex
      then begin
            MoveSize := (MaxIndex - Index) * SizeOf( TData);
            Move( fData[ Index + 1], fData[ Index], MoveSize)
           end;
     Dec( fCount);
     AfterRemoval( aData, Index);
     RemoveAt := aData
    end;
    
   function TVector.Replace( OldData, NewData: TData): SizeInt;
    var
     Index: SizeInt;
    begin
     Index := IndexOf( OldData);
     ReplaceAt( Index, NewData);
     Replace := Index
    end;
    
   function TVector.ReplaceAt( Index: SizeInt; NewData: TData): TData;
    var
     OldData: TData;
    begin
     Confirm( Index, False);
     if not CanRemove( fData[ Index], Index)
      then Error( 'custom removal error at ' + IntToStr( Index));
     if not CanInsert( NewData, Index)
      then Error( 'custom insertion error at ' + IntToStr( Index));
     OldData := fData[ Index];
     AfterRemoval( OldData, Index);
     BeforeInsertion( NewData, Index);
     fData[ Index] := NewData;
     ReplaceAt := OldData
    end;
    
   procedure TVector.SetDelta( NewDelta: SizeInt);
    begin
     if Delta < 0
      then Error( 'new Delta < 0');
     fDelta := NewDelta
    end;
    
   procedure TVector.SetLimit( NewLimit: SizeInt);
    begin
     if Limit < MaxIndex
      then Error( 'cold not set limit to ' + IntToStr( NewLimit));
     fLimit := NewLimit
    end;
    
   procedure TVector.Swap( Index1, Index2: SizeInt);
    var
     Temp: TData;
    begin
     Confirm( Index1, False);
     Confirm( Index2, False);
     Temp           := fData[ Index1];
     fData[ Index1] := fData[ Index2];
     fData[ Index2] := Temp
    end;
    
  procedure RegisterType( aTypeClass: TTypeClass);
   var
    Index: Integer;
    StreamEntry: TStreamEntry;
   begin
    if LastTypeClass <> aTypeClass
     then begin
           LastTypeClass := aTypeClass;
           StreamEntry := TStreamEntry.Create( aTypeClass);
           Index := Registry.Search( StreamEntry.Key);
           if Registry.Flag[ tfFound]
            then StreamEntry.Free
            else Registry.InsertAt( Index, StreamEntry)
          end
   end;
   
  procedure RegisterZClassesClasses;
   begin
    RegisterType( TBitAry);
    RegisterType( TData);
    RegisterType( TFloat);
    RegisterType( TFloatStack);
    RegisterType( TItem);
    RegisterType( TPool);
    RegisterType( TSequence);
    RegisterType( TStack);
    RegisterType( TVector)
   end;
   
  function ReadCardinal( S: TStream): Cardinal;
   begin
    S.ReadBuffer( Result, SizeOf( Cardinal))
   end;

  function ReadFloat( S: TStream): Float;
   begin
    S.ReadBuffer( Result, SizeOf( Float))
   end;

  function ReadInteger( S: TStream): Integer;
   begin
    S.ReadBuffer( Result, SizeOf( Integer))
   end;
   
  function ReadSizeInt( S: TStream): SizeInt;
   begin
    S.ReadBuffer( Result, SizeOf( SizeInt));
   end;

  function ReadType( aTypeClass: TTypeClass; S: TStream;
                     DataParent: TVector = nil): TType;
   var
    ClassName: string;
    Index: Integer;
   begin
    ClassName := S.ReadANSIString;
    if (aTypeClass <> nil) and (aTypeClass.ClassName <> ClassName)
     then raise EZtream.Create( 'wrong class (' + ClassName +
                                ') found instead of ' + aTypeClass.ClassName);
    Index := Registry.Search( @ClassName);
    if not Registry.Flag[ tfFound]
     then raise EZtream.Create( 'class ' + ClassName + ' unregistered');
    ReadType := TStreamEntry( Registry.Data[ Index]).CreateType;
    if Result.InheritsFrom( TData)
     then TData( Result).fParent := DataParent;
    Result.ReadFrom( S)
   end;

  procedure WriteCardinal( aCardinal: Cardinal; S: TStream);
   begin
    S.WriteBuffer( aCardinal, SizeOf( Cardinal))
   end;

  procedure WriteFloat( aFloat: Float; S: TStream);
   begin
    S.WriteBuffer( aFloat, SizeOf( Float))
   end;

  procedure WriteInteger( anInteger: Integer; S: TStream);
   begin
    S.WriteBuffer( anInteger, SizeOf( Integer))
   end;
   
  procedure WriteSizeInt( aSizeInt: SizeInt; S: TStream);
   begin
    S.WriteBuffer( aSizeInt, SizeOf( SizeInt))
   end;

  procedure WriteType( aType: TType; S: TStream);
   begin
    RegisterType( TTypeClass( aType.ClassType));
    S.WriteANSIString( aType.ClassName);
    aType.WriteTo( S)
   end;

  initialization
   Registry := TSequence.Create;

  finalization
   Registry.Free

end.
