unit Zexpress;

{Math term parser core
 Part of the z4free library package.

 For use with the Free Pascal Compiler (http://www.freepascal.org/) or the
 Lazarus project (http://www.lazarus.freepascal.org/)

 Copyright (C) 2007 Robert H. Mangl

 Zexpress is free software: you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
 Foundation, either version 3 of the License, or (at your option) any later
 version.

 Zexpress is distributed in the hope that it will be useful, but WITHOUT ANY
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with
 Zexpress.  If not, see <http://www.gnu.org/licenses/>}
 
 {$mode objfpc}{$H+}

 interface
 
  uses Math,
       Zett, ZClasses;
       
  type
  
   TIDStatus = (isOK,
                isRegistered, isUnregistered,
                isEmpty, isSpacey, isDigitHeaded,
                isLocked, isClosingBracket);
   TTermMode = (tmTest, tmCompile, tmInterpret);
       
   CPDatawell = function: Float;
   CPOperator = function( A, B: Float): Float;
   //CPFunction is defined in Zett

   TTermStatus = record
     Code: Integer;
     SrcPos: SizeInt;
     Arg1, Arg2: Float
    end;
    
   TTerm = class( TData)
     private
      fStatus: TTermStatus;
      fSource: string;
      fTarget: TVector;
      procedure ParseSource( Mode: TTermMode);
      property Source: string read fSource write fSource;
      property Target: TVector read fTarget write fTarget;
     public
      constructor Create( aSource: string; aMode: TTermMode);
      destructor Destroy; override;
      function Run: Float;
      property Status: TTermStatus read fStatus;
    end;
    
   ETermReg = class( EZett)
     private
      fIDStatus: TIDStatus;
     public
      constructor Create( anIDStatus: TIDStatus);
      property IDStatus: TIDStatus read fIDStatus;
    end;

  const
  
   tsOK                = 00;
   tsUnknownSymbol     = -1;
   tsValueExpected     = -2;
   tsOperationExpected = -3;
   tsNoTarget          =  1;
   tsTooManyResults    =  2;
   tsOverflow          =  3;
   tsDivisionByZero    =  4;
   tsIllegalQuantity   =  5;
   tsGeneric           =  6;
   tsNext              =  7;

  var

   DecimalSymbol : Char = ',';
   RaiseRegError : Boolean = False;

  procedure SetDecimalSymbol( aDecimalSymbol: Char);
  procedure TermError( aCode: Integer);
  
  function RegisterConstant( anID: string; aValue: Float): TIDStatus;
  function RegisterParameter( anID: string; aVariable: PFloat): TIDStatus;
  function RegisterDatawell( anID: string; aDatawell: CPDatawell): TIDStatus;
  function RegisterFunction( anID: string; aFunction: CPFunction): TIDStatus;
  function RegisterOperator( anID: string; anOperator: CPOperator;
                                           aPriority: Integer): TIDStatus;
  function RegisterBracket( anOpener, aCloser: string): TIDStatus;
  function Unregister( anID: string): TIDStatus;

  function RegisterBAO: TIDStatus;
  function RegisterLevel3: TIDStatus;
  function RegisterRandom: TIDStatus;

 implementation
 
  uses SysUtils;
  
  type
  
   TElementType =
    (etConstant, etLiteral, etParameter, etDatawell, etFunction, etBracketTerm,
     etOperator, etClosingBracket, etEndOfTerm);
     
   TElement = class( TData)
     private
      fID: string;
      fType: TElementType;
      fRefCount: Cardinal;
     protected
      function CompareWith( aKey: Pointer): Integer; override;
      function GetKey: Pointer; override;
      procedure OperateOn( Stack: TFloatStack; var Status: TTermStatus);
       virtual; abstract;
     public
      constructor Create( anID: string; anElementType: TElementType);
      property EType: TElementType read fType;
      property ID: string read fID;
      function IsLocked: Boolean;
      function IsValue: Boolean;
      function IsOperation: Boolean;
      procedure Lock;
      procedure Unlock;
    end;
    
   TConstant = class( TElement)
     private
      fValue: Float;
     protected
       procedure OperateOn( Stack: TFloatStack; var Status: TTermStatus);
        override;
     public
      constructor Create( anID: string; aValue: Float);
    end;

   TLiteral = class( TConstant)
     public
      constructor Create( aValue: Float);
    end;
    
   TParameter = class( TElement)
     private
      fVariable: PFloat;
     protected
      procedure OperateOn( Stack: TFloatStack; var Status: TTermStatus);
       override;
     public
      constructor Create( anID: string; aVariable: PFloat);
    end;
    
   TDatawell = class( TElement)
     private
      fDatawellCP: CPDatawell;
     protected
      procedure OperateOn( Stack: TFloatStack; var Status: TTermStatus);
       override;
     public
      constructor Create( anID: string; aDatawellCP: CPDatawell);
    end;
    
   TFunction = class( TElement)
     private
      fFunctionCP: CPFunction;
     protected
      procedure OperateOn( Stack: TFloatStack; var Status: TTermStatus);
       override;
     public
      constructor Create( anID: string; aFunctionCP: CPFunction);
    end;
    
   TOperation = class( TElement)
     protected
      function GetPriority: LongInt; virtual; abstract;
     public
      property Priority: LongInt read GetPriority;
    end;

   TOperator = class( TOperation)
     private
      fOperatorCP: CPOperator;
      fPriority: Integer;
     protected
      function GetPriority: LongInt; override;
      procedure OperateOn( Stack: TFloatStack; var Status: TTermStatus);
       override;
     public
      constructor Create( anID: string; aOperatorCP: CPOperator;
                                        aPriority: Integer);
    end;
    
   TEndOfTerm = class( TOperation)
     protected
      function GetPriority: LongInt; override;
     public
      constructor Create( anID: string);
    end;

   TBrackets = class( TElement)
     private
      fCloser: TEndOfTerm;
     public
      constructor Create( anOpener, aCloser: string);
      destructor Destroy; override;
      property Closer: TEndOfTerm read fCloser;
    end;
    
   TOpCode = class( TData)
     private
      fElement: TElement;
      fSrcPos: SizeInt;
     public
      constructor Create( anElement: TElement; aSrcPos: SizeInt);
      destructor Destroy; override;
      property Element: TElement read fElement;
      property SrcPos: SizeInt read fSrcPos;
    end;

   TRegistry = class( TSequence) // contains registered elements
     private
      function GetElementByID( anID: string): TElement;
     public
      constructor Create;
      destructor Destroy; override;
      property Element[ ID: string]: TElement read GetElementByID;
      function Identify( Source: string; SrcPos: SizeInt): TElement;
      function IsRegistered( anID: string): Boolean;
      function RegisterLiteral( aNumber: Float): TElement;
    end;
    
   ETerm = class( EZett)
     fCode: Integer;
     constructor Create( aCode: Integer);
    end;
    
  const
   Digits = ['0' .. '9'];
   IntegerSymbols = ['+', '-'] + Digits;
   ExponentialSymbols = ['e', 'E'];
   ValueTypes = [etConstant .. etBracketTerm];
   OperationTypes = [etOperator .. etEndOfTerm];

  var
   Registry : TRegistry = nil;
   EndOfLine : TEndOfTerm = nil;
   MantissaSymbols : set of Char = [];
   
  // ETerm methods
   
     constructor ETerm.Create( aCode: Integer);
      var
       Msg: string;
      begin
       fCode := aCode;
       Msg := 'term error. code ' + IntToStr( aCode);
       inherited Create( Msg)
      end;
      
  // ETermReg methods
       
     constructor ETermReg.Create( anIDStatus: TIDStatus);
     
      begin
       fIDStatus := anIDStatus;
       inherited Create( 'term registry error #' + IntToStr( Ord( anIDStatus)))
      end;
      
  // TBrackets methods
   
     constructor TBrackets.Create( anOpener, aCloser: string);
      begin
       inherited Create( anOpener, etBracketTerm);
       fCloser := TEndOfTerm.Create( aCloser);
       fCloser.fType := etClosingBracket
      end;

     destructor TBrackets.Destroy;
      begin
       Registry.FreeData( fCloser);
       inherited Destroy
      end;

  // TConstant methods
   
     constructor TConstant.Create( anID: string; aValue: Float);
      begin
       inherited Create( anID, etConstant);
       fValue :=aValue
      end;

     procedure TConstant.OperateOn( Stack: TFloatStack;
                                    var Status: TTermStatus);
      begin
       Stack.PushVal( fValue)
      end;
       
  // TDatawell methods
       
     constructor TDatawell.Create( anID: string; aDatawellCP: CPDatawell);
      begin
       inherited Create( anID, etDatawell);
       fDatawellCP := aDatawellCP
      end;
      
     procedure TDatawell.OperateOn( Stack: TFloatStack;
                                    var Status: TTermStatus);
      begin
       Stack.PushVal( fDatawellCP())
      end;

  // TElement methods

     constructor TElement.Create( anID: string; anElementType: TElementType);
      begin
       inherited Create;
       fID := anID;
       fType := anElementType;
       fRefCount := 0
      end;
      
     function TElement.CompareWith( aKey: Pointer): Integer;
      var
       KeyID: string;
      begin
       CompareWith := CompareStr( ID, PString( aKey)^);
      end;
      
     function TElement.GetKey: Pointer;
      begin
       GetKey := @fID
      end;
      
     function TElement.IsLocked: Boolean;
      begin
       IsLocked := fRefCount <> 0
      end;
      
     function TElement.IsValue: Boolean;
      begin
       IsValue := fType in ValueTypes
      end;
      
     function TElement.IsOperation: Boolean;
      begin
       IsOperation := fType in OperationTypes
      end;
      
     procedure TElement.Lock;
      begin
       if fRefCount < MaxCardinal
        then Inc( fRefCount);
      end;
      
     procedure TElement.Unlock;
      begin
       if fRefCount > 0
        then Dec( fRefCount)
      end;
      
  // TEndOfTerm methods

     constructor TEndOfTerm.Create( anID: string);
      begin
       inherited Create( anID, etEndOfTerm)
      end;

     function TEndOfTerm.GetPriority: LongInt;
      begin
       GetPriority := MaxInteger
      end;
      
  // TFunction methods
      
     constructor TFunction.Create( anID: string; aFunctionCP: CPFunction);
      begin
       inherited Create( anID, etFunction);
       fFunctionCP := aFunctionCP
      end;
      
     procedure TFunction.OperateOn( Stack: TFloatStack;
                                    var Status: TTermStatus);
      begin
       with Status do
        begin
         Arg1 := Stack.PullVal;
         Stack.PushVal( fFunctionCP( Arg1))
        end
      end;

  // TLiteral methods
  
     constructor TLiteral.Create( aValue: Float);
      var
       anID: string;
      begin
       Str( aValue, anID);
       inherited Create( anID, aValue);
       fType := etLiteral
      end;
      
  // TOPCode methods
      
     constructor TOpCode.Create( anElement: TElement; aSrcPos: SizeInt);
      begin
       inherited Create;
       fElement := anElement;
       fElement.Lock;
       fSrcPos := aSrcPos
      end;
      
     destructor TOpCode.Destroy;
      begin
       fElement.Unlock;
       inherited Destroy
      end;
      
  // TOperator methods

     constructor TOperator.Create( anID: string; aOperatorCP: CPOperator;
                                                 aPriority: Integer);
      begin
       inherited Create( anID, etOperator);
       fOperatorCP := aOperatorCP;
       if aPriority = MaxInteger
        then fPriority := MaxInteger - 1
        else if aPriority = MinInteger
              then fPriority := MinInteger + 1
              else fPriority := aPriority
      end;

     function TOperator.GetPriority: LongInt;
      begin
       GetPriority := fPriority
      end;
      
     procedure TOperator.OperateOn( Stack: TFloatStack;
                                    var Status: TTermStatus);
      var
       F: TFloat;
      begin
       with Status do
        begin
         Arg2 := Stack.PullVal;
         Arg1 := Stack.PullVal;
         Stack.PushVal( fOperatorCP( Arg1, Arg2))
        end
      end;

  // TParameter methods

     constructor TParameter.Create( anID: string; aVariable: PFloat);
      begin
       inherited Create( anID, etParameter);
       fVariable := aVariable
      end;
      
     procedure TParameter.OperateOn( Stack: TFloatStack;
                                     var Status: TTermStatus);
      begin
       Stack.PushVal( fVariable^)
      end;

  // TRegistry methods
      
     constructor TRegistry.Create;
      begin
       inherited Create;
       Flag[ tfReversed] := True
      end;

     destructor TRegistry.Destroy;
      var
       I, c: SizeInt;
       E: TElement;
       EType: TElementType;
      begin
       if Count > 0
        then begin
              I := 0;
              repeat
               E := TElement( Data[ I]);
               EType := E.EType;
               if EType = etClosingBracket
                then Inc( I)
                else begin
                  FreeAt( I);
                  if EType = etBracketTerm
                   then I := 0
                 end;
              until Count = 0
             end;
       inherited Destroy
      end;
      
     function TRegistry.GetElementByID( anID: string): TElement;
      var
       Index: Integer;
      begin
       Index := Search( @anID);
       if Flag[ tfFound]
        then GetElementByID := TElement( Data[ Index])
        else GetElementByID := nil
      end;
     
     function TRegistry.Identify( Source: string; SrcPos: SizeInt): TElement;
      var
       Elem: TElement;
       I, IDLen: SizeInt;
       ID, Tested: string;
      begin
       Identify := nil;
       if (SrcPos > Length( Source)) or (Count = 0)
        then Exit;
       for I := 0 to Count - 1 do
        begin
         Elem := TElement( Data[ I]);
         ID := Elem.ID;
         IDLen := Length( ID);
         Tested := Copy( Source, SrcPos, IDLen);
         if Copy( Source, SrcPos, IDLen) = ID
          then begin
                Identify := Elem;
                Break
               end
        end
      end;
      
     function TRegistry.IsRegistered( anID: string): Boolean;
      begin
       IsRegistered := Element[ anID] <> nil
      end;

     function TRegistry.RegisterLiteral( aNumber: Float): TElement;
      var
       Elem: TElement;
       LitStr: string;
      begin
       Str( aNumber, LitStr);
       Elem := Element[ LitStr];
       if Elem = nil
        then begin
              Elem := TLiteral.Create( aNumber);
              Insert( Elem)
             end;
       RegisterLiteral := Elem
      end;
      
  // TTerm methods

     constructor TTerm.Create( aSource: string; aMode: TTermMode);
      begin
       inherited Create;
       if aMode = tmCompile
        then fTarget := TVector.Create
        else fTarget := nil;
       fSource := aSource;
       ParseSource( aMode);
      end;

     destructor TTerm.Destroy;
      begin
       fTarget.Free;
       inherited Destroy
      end;

     procedure TTerm.ParseSource( Mode: TTermMode);
      var
       Stack: TFloatStack;
       Operation: TOperation;
      procedure Parse( var SrcPos: SizeInt; EndOfTerm: TEndOfTerm); // Parse
       var
        OpStack: TStack;
        Element: TElement;
       procedure Handle( Element: TElement); // Handle
        begin
         if not( Element.EType in [etBracketTerm, etClosingBracket])
          then begin
                case Mode of
                  tmCompile: Target.Insert( TOpCode.Create( Element, SrcPos));
                  tmInterpret: Element.OperateOn( Stack, fStatus)
                 end
               end
        end;
       function LiteralGet: TElement; // LiteralGet
        var
         Lit: Float; // the literal as number
         LitStr: string; // the literal as string
         ErrPos: SizeInt; // position of first non-literal string
         PosInc: SizeInt; // increment to go to the next symbol
        begin
         LitStr := FromStr( Source, SrcPos);
         Val( LitStr, Lit, ErrPos);
         if ErrPos <> 0
          then begin
                PosInc := ErrPos - 1;
                LitStr := Copy( LitStr, 1, PosInc);
                Val( LitStr, Lit, ErrPos)
               end
          else PosInc := Length( LitStr);
         Inc( SrcPos, PosInc);
         LiteralGet := Registry.RegisterLiteral( Lit)
        end;
       function ElementGet: TElement; // ElementGet
        var
         SourceLen: SizeInt;
         Element: TElement;
        begin
         SourceLen := Length( Source);
         while (SrcPos <= SourceLen) and (Source[ SrcPos] = #032) do
          Inc( SrcPos);
         if SrcPos > SourceLen
          then ElementGet := EndOfLine
          else begin
                Element := Registry.Identify( Source, SrcPos);
                if Element = nil
                 then if Source[ SrcPos] in MantissaSymbols
                       then ElementGet := LiteralGet
                       else TermError( tsUnknownSymbol)
                 else begin
                       ElementGet := Element;
                       Inc( SrcPos, Length( Element.ID))
                      end
               end;
        end;
       procedure ParseArgument; forward;
       function ValueGet: TElement; // ValueGet
        var
         Element: TElement;
        begin
         Element := ElementGet;
         if not Element.IsValue
          then TermError( tsValueExpected)
          else begin
                case Element.EType of
                   etFunction   : ParseArgument;
                   etBracketTerm: Parse( SrcPos, TBrackets( Element).Closer)
                  end;
                ValueGet := Element
               end
        end;
       procedure ParseArgument; // ParseArgument
        begin
         Handle( ValueGet)
        end;
       function OperationGet: TOperation; // OperationGet
        var
         Element: TElement;
        begin
         Element := ElementGet;
         if not Element.IsOperation
          then TermError( tsOperationExpected)
          else OperationGet := TOperation( Element)
        end;
       function TopPriority: Integer; // TopPriority
        var
         Top: TOperation;
        begin
         Top := TOperation( OpStack.ToS);
         if Top = nil
          then TopPriority := MinInteger
          else TopPriority := Top.Priority
        end;
       begin // Parse
        OpStack := TStack.Create;
        repeat
         Handle( ValueGet);
         Operation := OperationGet;
         if Operation = EndOfTerm
          then Break;
         while TopPriority >= Operation.Priority do
          Handle( TElement( OpStack.Pull));
         OpStack.Push( Operation)
        until False;
        while OpStack.ToS <> nil do
         Handle( TElement( OpStack.Pull));
        OpStack.Free
       end;
      begin // ParseSource
       Source := ReplaceChar( Source, DecimalSymbol, '.');
       case Mode of
         tmTest     : begin
                       Target := nil;
                       Stack := nil
                      end;
         tmCompile  : begin
                       Target := TVector.Create;
                       Stack := nil
                      end;
         tmInterpret: begin
                       Target := nil;
                       Stack := TFloatStack.Create
                      end
        end;
       with Status do
        begin
         Code := tsOK;
         SrcPos := 1;
         Arg1 := 0;
         Arg2 := 0;
         try
            Parse( SrcPos, EndOfLine);
            if Mode = tmInterpret
             then begin
                   if Stack.Count > 1
                    then TermError( tsTooManyResults);
                   Arg1 := Stack.PullVal;
                  end
           except
            on EOverflow        do Code := tsOverflow;
            on ETerm            do Code := ETerm( CurrentException).fCode;
            on EInvalidArgument do Code := tsIllegalQuantity
            else                   Code := tsGeneric
          end
        end;
       Stack.Free
      end;
     
     function TTerm.Run: Float;
      var
       I: SizeInt;
       OpCode: TOpCode;
       Stack: TFloatStack;
      begin
       with Status do
        begin
         SrcPos := 1;
         Arg1 := 0;
         Arg2 := 0;
         if (Target = nil) or (Target.Count = 0)
          then Code := tsNoTarget
          else begin
                Code := tsOK;
                Stack := TFloatStack.Create;
                try
                  for I := 0 to Target.Count - 1 do
                   begin
                    OpCode := TOPCode( Target.Data[ I]);
                    OpCode.Element.OperateOn( Stack, Status)
                   end;
                   if Stack.Count > 1
                    then TermError( tsTooManyResults);
                   Run := Stack.PullVal;
                  except
                   on EOverflow        do begin
                                        Code := tsOverflow;
                                        SrcPos := OpCode.SrcPos
                                       end;
                   on ETerm            do begin
                                        Code := ETerm( CurrentException).fCode;
                                        SrcPos := OpCode.SrcPos
                                       end;
                   on EInvalidArgument do begin
                                           Code := tsIllegalQuantity;
                                           SrcPos := OpCode.SrcPos
                                          end
                   else                begin
                                        Code := tsGeneric;
                                        SrcPos := OpCode.SrcPos
                                       end
                 end;
                Stack.Free
               end
        end
      end;
      
  function DoPower( A: Float; B: TRational): Float; forward;
      
  // term symbol implementations
  
     // datawells
  
     function RndTSI: Float;
      begin
       RndTSI := Random
      end;
      
     // functions
  
     function AbsTSI( X: Float): Float;
      begin
       Result := -Abs( X)
      end;

     function ChsTSI( X: Float): Float;
      begin
       Result := -X
      end;
      
     function LdTSI( X: Float): Float;
      begin
       Result := Ln( X) / Ln( 2)
      end;

     function LgTSI( X: Float): Float;
      begin
       Result := Ln( X) / Ln( 10)
      end;

     function LnTSI( X: Float): Float;
      begin
       Result := Ln( X)
      end;

     function SgnTSI( X: Float): Float;
      begin
       Result := -Sign( X)
      end;

     // operators
      
     function AddTSI( A, B: Float): Float;
      begin
       Result := A + B
      end;

     function DieRollTSI( A, B: Float): Float;
      var
       I, N, S: Cardinal;
      begin
       if (Frac( A) <> 0) or (Frac( B) <> 0)
        then TermError( tsIllegalQuantity);
       if (A < 0) or (B < 1)
        then TermError( tsIllegalQuantity);
       Result := 0;
       if A = 0
        then Exit;
       N := Trunc( A);
       S := Trunc( B);
       for I := 1 to N do
        Result := Result + Int( RndTSI * S + 1)
      end;
      
     function DivTSI( A, B: Float): Float;
      begin
       if B = 0
        then TermError( tsDivisionByZero);
       Result := A / B
      end;
      
     function LogTSI( A, B: Float): Float;
      begin
       Result := Ln( B) / Ln( A)
      end;

     function MulTSI( A, B: Float): Float;
      begin
       Result := A * B
      end;

     function PowerTSI( A, B: Float): Float;
      begin
       PowerTSI := DoPower( A, Rationalize( B))
      end;
      
     function RootTSI( A, B: Float): Float;
      var
       QOld: Integer;
       BRational: TRational;
      begin
       if B = 0
        then TermError( tsIllegalQuantity);
       BRational := Rationalize( B);
       with BRational do
        if IsRational
         then begin
               QOld := Q;
               Q := Trunc( I) * Q + P;
               P := QOld;
               I := 0;
               CancelDown( P, Q);
               if Q < 0
                then begin
                      Q := -Q;
                      P := -P
                     end
              end
         else I := 1 / I;
       RootTSI := DoPower( A, BRational)
      end;

     function SubTSI( A, B: Float): Float;
      begin
       Result := A - B
      end;
      
  function DoPower( A: Float; B: TRational): Float;
   begin
    with B do
     if A < 0
      then if IsRational
            then if Odd( Q)
                  then Result := Power(A, I) * ( 1 - 2 * Ord( Odd( P)))
                  else TermError( tsIllegalQuantity)
            else TermError( tsIllegalQuantity)
      else begin
            if IsRational
             then I := I + 1.0 * P / Q;
            if A = 0
             then Result := Ord( I = 0)
             else Result := Power( A, I)
           end
   end;

  procedure SetDecimalSymbol( aDecimalSymbol: Char);
   begin
    if aDecimalSymbol in ( IntegerSymbols + ExponentialSymbols)
     then raise EZett.Create( 'illegal decimal symbol')
     else begin
           DecimalSymbol := aDecimalSymbol;
           MantissaSymbols := IntegerSymbols + [DecimalSymbol]
          end
   end;
   
  function TestID( anID: string): TIDStatus;
   begin
    if anID = ''
     then TestID := isEmpty
     else if Pos( ' ', anID) <> 0
           then TestID := isSpacey
           else if anID[ 1] in Digits
                 then TestID := isDigitHeaded
                 else if Registry.IsRegistered( anID)
                       then TestID := isRegistered
                       else TestID := isUnregistered
   end;
   
  function RegisterBAO: TIDStatus; // basic arithmetical operations
   var
    OldRaiseRegError: Boolean;
   begin
    OldRaiseRegError := RaiseRegError;
    RaiseRegError := True;
    RegisterBAO := isOK;
    try
       RegisterBracket( '(', ')');
       RegisterFunction( 'chs', @ChsTSI);
       RegisterFunction( 'abs', @AbsTSI);
       RegisterFunction( 'sgn', @SgnTSI);
       RegisterOperator( '+', @AddTSI,   0);
       RegisterOperator( '-', @SubTSI,   0);
       RegisterOperator( '*', @MulTSI, 100);
       RegisterOperator( '/', @DivTSI, 100);
      except
       on ETermReg do RegisterBAO := ETermReg( CurrentException).IDStatus
     end;
    if OldRaiseRegError
     then ETermReg.Create( Result)
     else RaiseRegError := OldRaiseRegError
   end;

  function RegisterBracket( anOpener, aCloser: string): TIDStatus;
   var
    TestResult: TIDStatus;
    Brackets: TBrackets;
   begin
    TestResult := TestID( anOpener);
    if TestResult = isUnregistered
     then begin
           TestResult := TestID( aCloser);
           if TestResult = isUnregistered
            then begin
                  Brackets := TBrackets.Create( anOpener, aCloser);
                  Registry.Insert( Brackets);
                  Registry.Insert( Brackets.fCloser);
                  TestResult := isOK
                 end
          end;
    RegisterBracket := TestResult;
    if RaiseRegError and (Result <> isOK)
     then raise ETermReg.Create( Result)
   end;

  function RegisterConstant( anID: string; aValue: Float): TIDStatus;
   var
    TestResult: TIDStatus;
   begin
    TestResult := TestID( anID);
    if TestResult = isUnregistered
     then begin
           Registry.Insert( TConstant.Create( anID, aValue));
           RegisterConstant := isOK
          end
     else RegisterConstant := TestResult;
    if RaiseRegError and (Result <> isOK)
     then raise ETermReg.Create( Result)
   end;
   
  function RegisterDatawell( anID: string; aDatawell: CPDatawell): TIDStatus;
   var
    TestResult: TIDStatus;
   begin
    TestResult := TestID( anID);
    if TestResult = isUnregistered
     then begin
           Registry.Insert( TDatawell.Create( anID, aDatawell));
           RegisterDatawell := isOK
          end
     else RegisterDatawell := TestResult;
    if RaiseRegError and (Result <> isOK)
     then raise ETermReg.Create( Result)
   end;

  function RegisterFunction( anID: string; aFunction: CPFunction): TIDStatus;
   var
    TestResult: TIDStatus;
   begin
    TestResult := TestID( anID);
    if TestResult = isUnregistered
     then begin
           Registry.Insert( TFunction.Create( anID, aFunction));
           RegisterFunction := isOK
          end
     else RegisterFunction := TestResult;
    if RaiseRegError and (Result <> isOK)
     then raise ETermReg.Create( Result)
   end;

  function RegisterLevel3: TIDStatus;
   var
    OldRaiseRegError: Boolean;
   begin
    OldRaiseRegError := RaiseRegError;
    RaiseRegError := True;
    RegisterLevel3 := isOK;
    try
       RegisterFunction( 'ld', @LdTSI);
       RegisterFunction( 'ln', @LnTSI);
       RegisterFunction( 'lg', @LgTSI);
       RegisterOperator( '^', @PowerTSI, 200);
       RegisterOperator( 'rdx', @RootTSI, 200);
       RegisterOperator( 'log', @LogTSI, (MaxInteger div 100) * 100);
      except
       on ETermReg do RegisterLevel3 := ETermReg( CurrentException).IDStatus
     end;
    Randomize;
    if OldRaiseRegError
     then ETermReg.Create( Result)
     else RaiseRegError := OldRaiseRegError
   end;

  function RegisterOperator( anID: string; anOperator: CPOperator;
                                           aPriority: Integer): TIDStatus;
   var
    TestResult: TIDStatus;
   begin
    TestResult := TestID( anID);
    if TestResult = isUnregistered
     then begin
           Registry.Insert( TOperator.Create( anID, anOperator, aPriority));
           RegisterOperator := isOK
          end
     else RegisterOperator := TestResult;
    if RaiseRegError and (Result <> isOK)
     then raise ETermReg.Create( Result)
   end;

  function RegisterParameter( anID: string; aVariable: PFloat): TIDStatus;
   var
    TestResult: TIDStatus;
   begin
    TestResult := TestID( anID);
    if TestResult = isUnregistered
     then begin
           Registry.Insert( TParameter.Create( anID, aVariable));
           RegisterParameter := isOK
          end
     else RegisterParameter := TestResult;
    if RaiseRegError and (Result <> isOK)
     then raise ETermReg.Create( Result)
   end;

  function RegisterRandom: TIDStatus;
   var
    OldRaiseRegError: Boolean;
   begin
    OldRaiseRegError := RaiseRegError;
    RaiseRegError := True;
    RegisterRandom := isOK;
    try
       RegisterDatawell( 'rnd', @RndTSI);
       RegisterOperator( 'd', @DieRollTSI, MaxInteger);
      except
       on ETermReg do RegisterRandom := ETermReg( CurrentException).IDStatus
     end;
    Randomize;
    if OldRaiseRegError
     then ETermReg.Create( Result)
     else RaiseRegError := OldRaiseRegError
   end;

  procedure TermError( aCode: Integer);
   begin
    raise ETerm.Create( aCode)
   end;
   
  function Unregister( anID: string): TIDStatus;
   var
    TestResult: TIDStatus;
    Element: TElement;
   begin
    TestResult := TestID( anID);
    if TestResult = isRegistered
     then begin
           Element := Registry.GetElementByID( anID);
           if Element.IsLocked
            then Unregister := isLocked
            else if Element.EType = etClosingBracket
                  then Unregister := isClosingBracket
                  else begin
                        Registry.FreeData( Element);
                        Unregister := isOK
                       end
          end
     else Unregister := TestResult;
    if RaiseRegError and (Result <> isOK)
     then raise ETermReg.Create( Result)
   end;
   
  initialization
   Registry := TRegistry.Create;
   EndOfLine := TEndOfTerm.Create( '');
   MantissaSymbols := IntegerSymbols + [DecimalSymbol];

  finalization
   EndOfLine.Free;
   Registry.Free
      
end.

